/*
cat /dev/urandom|ffmpeg -pix_fmt rgb24 -f rawvideo -s 200x100 -r 25 -i - toto.mpg
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include "png.h"

/* transition types */
#define MIX  0
#define SWAP 1

float transition_time = 100;
float exposition_time = 1000;
float framerate = 25;
int width, height;
char *img0, *img1, *buf;

int transition_type = MIX;

char *oneline(FILE *f)
{
#define PUT(x) do { \
    len++; \
    l = realloc(l, len); if (l==NULL) abort(); \
    l[len-1] = x; \
  } while (0)

  static char *l = NULL;
  int len;
  int c;

  len = 0;

  while (1) {
    c = fgetc(f); if (c == '\n' || c == EOF) break;
    PUT(c);
  }

  if (c == EOF && len == 0) return NULL;

  PUT(0);
  return l;
}

int load_image(char *l)
{
  pngdata p;
  while (isspace(*l) && *l) l++;
  if (access(l, R_OK))
    { fprintf(stderr, "ERROR: cannot access file '%s'\n", l); return 1; }
  if (load_png(&p, l))
    { fprintf(stderr, "ERROR: load_png failed\n"); return 1; }
  if (width == 0) { width = p.width; height = p.height; }
  if (width != p.width || height != p.height)
    { fprintf(stderr, "ERROR: all images must have same size\n"); return 1; }
  img1 = p.data;
  return 0;
}

void expose_image(void)
{
  int i;
  for (i = 0; i < (float)framerate * exposition_time / 1000.; i++)
    fwrite(img0, width*height*3, 1, stdout);
}

void expose_transition(void)
{
  int i, j;
  int frames = (float)framerate * transition_time / 1000.;
  if (buf == NULL) { buf = malloc(width*height*3); if (buf == NULL) abort(); }
  switch (transition_type) {
  case MIX:
    for (i = 0; i < frames; i++) {
      float q = (float)i / frames;
      for (j = 0; j < width*height*3; j++)
        buf[j] = (1-q) * (unsigned char)img0[j] + q * (unsigned char)img1[j];
      fwrite(buf, width*height*3, 1, stdout);
    }
    break;
  case SWAP:
    for (i = 0; i < frames; i++) {
      int frontier = (float)(frames-i) / frames * width;
      int l;
      for (l = 0; l < height; l++) {
        memcpy(buf + l*width*3,
               img0 + l*width*3 + (width-frontier)*3,
               frontier*3);
        memcpy(buf + l * width*3 + frontier*3,
               img1 + l*width*3,
               (width-frontier)*3);
      }
      fwrite(buf, width*height*3, 1, stdout);
    }
    break;
  }
}

int main(int n, char **v)
{
  FILE *f;

  if (n != 2) {
    fprintf(stderr, "gimme a program file\n");
    fprintf(stderr, "lines of it are:\n");
    fprintf(stderr, "    T <transition time (ms) (default 100.)>\n");
    fprintf(stderr, "    E <exposition time (ms) (default 1000.)>\n");
    fprintf(stderr, "    F <framerate (default 25.)>\n");
    fprintf(stderr, "    I <filename of a PNG image>\n");
    fprintf(stderr,
        "    [Sswap | Smix]: transition type, swap or mix (default mix)\n");
    fprintf(stderr, "all images must have same width/height\n");
    return 1;
  }

  f = fopen(v[1], "r");
  if (f == NULL) { perror(v[1]); return 1; }

  while (1) {
    char *l = oneline(f);
    if (l==NULL) break;
    fprintf(stderr, "INFO: processing '%s'\n", l);
    switch (l[0]) {
    default: fprintf(stderr, "WARNING: unhandled line '%s'\n", l); break;
    case 0: break;
    case 'T': if (sscanf(l+1, "%g", &transition_time) != 1) goto err; break;
    case 'E': if (sscanf(l+1, "%g", &exposition_time) != 1) goto err; break;
    case 'F': if (sscanf(l+1, "%g", &framerate) != 1) goto err; break;
    case 'I':
      if (load_image(l+1)) goto err;
      if (img0) { expose_image(); expose_transition(); }
      free(img0);
      img0 = img1;
      break;
    case 'S':
      if (!strcmp(l, "Sswap")) transition_type = SWAP;
      else if (!strcmp(l, "Smix")) transition_type = MIX;
      break;
    }
  }

  if (img0) expose_image();

  fflush(stdout);

  return 0;

err:
  fprintf(stderr, "ERROR: some error!\n"); return 1;
}
